@@ -1,14 +1,31 @@
 from __future__ import annotations
 
 from functools import cached_property
-from typing import TYPE_CHECKING, Callable, NewType, Optional, Type, TypedDict, TypeVar, Union
+from typing import (
+    TYPE_CHECKING,
+    List,
+    Optional,
+    TypedDict,
+    Union,
+)
 
 import graphql
 from graphql import OperationType
-from graphql.type import definition as gql_def
+from graphql import language as gql_lang
+from graphql.language import visitor
 
+from qtgql.codegen.graphql_ref import (
+    is_enum_definition,
+    is_list_definition,
+    is_non_null_definition,
+    is_object_definition,
+    is_operation_def_node,
+    is_scalar_definition,
+    is_selection_set,
+    is_union_definition,
+)
 from qtgql.codegen.py.compiler.builtin_scalars import BuiltinScalars
-from qtgql.codegen.py.compiler.query import QueryHandlerDefinition
+from qtgql.codegen.py.compiler.query import QtGqlQueriedField, QtGqlQueryHandlerDefinition
 from qtgql.codegen.py.compiler.template import (
     TemplateContext,
     handlers_template,
@@ -26,37 +43,95 @@ from qtgql.codegen.utils import anti_forward_ref
 from qtgql.exceptions import QtGqlException
 
 if TYPE_CHECKING:  # pragma: no cover
+    from graphql.type import definition as gql_def
+
     from qtgql.codegen.py.compiler.config import QtGqlConfig
 
 introspection_query = graphql.get_introspection_query(descriptions=True)
 
 
-OperationName = NewType("OperationName", str)
-
-
 class GeneratedNamespace(TypedDict):
     handlers: str
     objecttypes: str
 
 
+def has_id_field(selection_set: gql_lang.SelectionSetNode) -> bool:
+    if hasattr(selection_set, "__has_id__"):
+        return True
+    for field in selection_set.selections:
+        assert isinstance(field, gql_lang.FieldNode)
+        if field.name.value == "id":
+            return True
+    return False
+
+
+class IDInjectionVisitor(visitor.Visitor):
+    ID_SELECTION_NODE = gql_lang.FieldNode(
+        name=gql_lang.NameNode(value="id"), arguments=(), directives=()
+    )
+
+    def __init__(self) -> None:
+        super().__init__()
+        self.modified_source: str = ""
+
+    def enter_field(self, node: gql_def.FieldNode, key, parent, path, ancestors):
+        # root field grand parent is an operation.
+        if is_operation_def_node(ancestors[-2]):
+            return
+        selection_set: gql_lang.SelectionSetNode = ancestors[-1]
+        assert is_selection_set(selection_set)
+        if not has_id_field(selection_set):
+            selection_set.selections = (self.ID_SELECTION_NODE, *selection_set.selections)
+            selection_set.__has_id__ = True
+
+
+class OperationMinerVisitor(visitor.Visitor):
+    """Creates handlers for root operations."""
+
+    def __init__(self, evaluator: SchemaEvaluator):
+        super().__init__()
+        self.query_handlers: dict[str, QtGqlQueryHandlerDefinition] = {}
+        self.evaluator = evaluator
+
+    def enter_operation_definition(self, node, key, parent, path, ancestors):
+        if operation := is_operation_def_node(node):
+            if operation.operation is OperationType.QUERY:
+                root_field: gql_lang.FieldNode = operation.selection_set.selections[0]  # type: ignore
+                fname = root_field.name.value
+                assert self.evaluator._query_type
+                root_qtgql_field = QtGqlQueriedField.from_field(
+                    self.evaluator._query_type.fields_dict[fname], root_field.selection_set
+                )
+                op_name = operation.name.value
+                self.query_handlers[op_name] = QtGqlQueryHandlerDefinition(
+                    query=graphql.print_ast(node),
+                    name=op_name,
+                    field=root_qtgql_field,
+                    directives=node.directives,
+                )
+
+
 class SchemaEvaluator:
     def __init__(self, config: QtGqlConfig):
         self.config = config
         self._generated_types: dict[str, GqlTypeDefinition] = {}
         self._generated_enums: EnumMap = {}
         self._fragments_store: dict[int, str] = {}
-        self._query_handlers: dict[OperationName, QueryHandlerDefinition] = {}
+        self._query_handlers: dict[str, QtGqlQueryHandlerDefinition] = {}
+        self._query_type: Optional[GqlTypeDefinition] = None
 
     @cached_property
     def schema_definition(self) -> graphql.GraphQLSchema:
-        with open(self.config.graphql_dir / "schema.graphql", "r") as f:
+        with (self.config.graphql_dir / "schema.graphql").open() as f:
             return graphql.build_schema(f.read())
 
-    @property
-    def query_type(self) -> GqlTypeDefinition:
-        query_type = self.schema_definition.query_type
-        assert query_type
-        return self._generated_types[query_type.name]
+    @cached_property
+    def root_types(self) -> List[Optional[gql_def.GraphQLObjectType]]:
+        return [
+            self.schema_definition.get_root_type(OperationType.QUERY),
+            self.schema_definition.get_root_type(OperationType.MUTATION),
+            self.schema_definition.get_root_type(OperationType.SUBSCRIPTION),
+        ]
 
     def _evaluate_field_type(self, t: gql_def.GraphQLType) -> GqlTypeHinter:
         if non_null := is_non_null_definition(t):
@@ -112,13 +187,31 @@ class SchemaEvaluator:
             return None
         if evaluated := self._generated_types.get(t_name, None):
             return evaluated
-
-        concrete = GqlTypeDefinition(
+        if type_ not in self.root_types:
+            try:
+                id_field = type_.fields["id"]
+                if nonull := is_non_null_definition(id_field.type):
+                    id_scalar = graphql.type.scalars.GraphQLID
+                    if nonull.of_type is not id_scalar:
+                        raise QtGqlException(
+                            f"id field type must be of the {id_scalar} scalar!"
+                            f"\n Got: {nonull.of_type}"
+                        )
+                else:
+                    raise QtGqlException("id field should not be nullable!" f"\n Got: {id_field}")
+            except KeyError:
+                raise QtGqlException(
+                    "QtGql enforces types to have ID field"
+                    f"type {type_} does not not define an id field.\n"
+                    f"fields: {type_.fields}"
+                )
+        return GqlTypeDefinition(
             name=t_name,
             docstring=type_.description,
-            fields=[self._evaluate_field(name, field) for name, field in type_.fields.items()],
+            fields_dict={
+                name: self._evaluate_field(name, field) for name, field in type_.fields.items()
+            },
         )
-        return concrete
 
     def _evaluate_enum(self, enum: gql_def.GraphQLEnumType) -> Optional[GqlEnumDefinition]:
         name: str = enum.name
@@ -138,41 +231,35 @@ class SchemaEvaluator:
         for name, type_ in self.schema_definition.type_map.items():
             if name.startswith("__"):
                 continue
-
             if object_definition := is_object_definition(type_):
                 if object_type := self._evaluate_object_type(object_definition):
+                    if object_definition is self.schema_definition.query_type:
+                        self._query_type = object_type
+                    assert object_type not in self.root_types
+
                     self._generated_types[object_type.name] = object_type
             elif enum_def := is_enum_definition(type_):
                 if enum := self._evaluate_enum(enum_def):
                     self._generated_enums[enum.name] = enum
 
-    def parse_queries(self) -> None:
-        with open(self.config.graphql_dir / "operations.graphql", "r") as f:
-            queries = graphql.parse(f.read())
+    def parse_operations(self) -> None:
+        with (self.config.graphql_dir / "operations.graphql").open() as f:
+            operations = graphql.parse(f.read())
 
-        if errors := graphql.validate(self.schema_definition, queries):
+        operations = visitor.visit(operations, IDInjectionVisitor())
+        # validate the operation against the static schema
+        if errors := graphql.validate(self.schema_definition, operations):
             raise QtGqlException([error.formatted for error in errors])
 
-        for definition in queries.definitions:
-            field_name = definition.selection_set.selections[0].name.value  # type: ignore
-            field = None
-            if definition.operation is OperationType.QUERY:  # type: ignore
-                for field_impl in self.query_type.fields:
-                    if field_impl.name == field_name:
-                        field = field_impl
-                assert field
-                op_name = OperationName(definition.name.value)  # type: ignore
-                self._query_handlers[op_name] = QueryHandlerDefinition(
-                    query=graphql.print_ast(definition),
-                    name=op_name,
-                    field=field,
-                    directives=definition.directives,  # type: ignore
-                )
+        # get QtGql fields from the query AST.
+        operation_miner = OperationMinerVisitor(self)
+        visitor.visit(operations, operation_miner)
+        self._query_handlers.update(operation_miner.query_handlers)
 
     def dumps(self) -> GeneratedNamespace:
         """:return: The generated schema module as a string."""
         self.parse_schema_concretes()
-        self.parse_queries()
+        self.parse_operations()
         context = TemplateContext(
             enums=list(self._generated_enums.values()),
             types=[
@@ -188,27 +275,5 @@ class SchemaEvaluator:
     def dump(self):
         """:param file: Path to the directory the codegen would dump to."""
         for fname, content in self.dumps().items():
-            with open(self.config.graphql_dir / (fname + ".py"), "w") as fh:
+            with (self.config.graphql_dir / (fname + ".py")).open("w") as fh:
                 fh.write(content)
-
-
-T_Definition = TypeVar("T_Definition", bound=gql_def.GraphQLType)
-
-
-def definition_identifier_factory(
-    expected: Type[T_Definition],
-) -> Callable[[gql_def.GraphQLType], Optional[T_Definition]]:
-    def type_guarder(definition: gql_def.GraphQLType) -> Optional[T_Definition]:
-        if isinstance(definition, expected):
-            return definition
-
-    return type_guarder
-
-
-is_object_definition = definition_identifier_factory(gql_def.GraphQLObjectType)
-is_enum_definition = definition_identifier_factory(gql_def.GraphQLEnumType)
-is_list_definition = definition_identifier_factory(gql_def.GraphQLList)
-is_scalar_definition = definition_identifier_factory(gql_def.GraphQLScalarType)
-is_union_definition = definition_identifier_factory(gql_def.GraphQLUnionType)
-
-is_non_null_definition = definition_identifier_factory(gql_def.GraphQLNonNull)
